;Autor     : Erik Buchmann
;Projekt   : Multitasking
; Quarz    : 24 MHz
; Datum    : 4.10.100
;---------------------------------------------------------------

$NOMOD51
$INCLUDE (89C1051.MCU)

; Speicherbelegung 
;---------------------------------------------------------------
	TS_CURRTASK	DATA	 8	; Aktueller Task
	TS_R0BACKUP	DATA	 9	; Backup fr R0
	TS_ACCBACKUP	DATA	10	; Backup fr Acc
	TS_DPLBACKUP	DATA	11	; Backup fr DPTR
	TS_DPHBACKUP	DATA	12
	TS_BBACKUP	DATA	13	; Backup fr B
	
	TS_DATA		DATA	14	; Beginn des Speicherfeldes mit den SP

	TS_SP_TASK1	DATA	20	; Startadresse des Stackpointers Task1
	TS_SP_TASK2	DATA	40
	TS_SP_TASK3	DATA	60
	TS_SP_TASK4	DATA	80
	TS_SP_TASK5	DATA	100

; Ersatzvariablen
;---------------------------------------------------------------
	; Informationen ber die Taskumschaltung
	TS_TASKCOUNT	EQU	5	; Taskanzahl

	TS_STARTJMPVCT	EQU	20	; Anfang im Sprungvector
	TS_SPACE4PANDP	EQU	8	; Abstand im Sprungvector der Push's und Pops

org 0h
        ajmp start

; Interruptroutinen 
;---------------------------------------------------------------
org 0Bh
        ajmp timer0_int

; Funktionen 
;---------------------------------------------------------------
; Tabellen der push's und pop's
;  jeder der Eintrge ist TS_SPACE4PANDP Bytes lang - Lnge beachten, da
;  indirekt ber Tabelle indiziert wird!
org TS_STARTJMPVCT + (TS_SPACE4PANDP*0) ; Task1
	push 0
	jmp TS_switching_pushready
org TS_STARTJMPVCT + (TS_SPACE4PANDP*1)	; Task2
	push Acc
	jmp TS_switching_pushready
org TS_STARTJMPVCT + (TS_SPACE4PANDP*2)	; Task3
	push Acc
	jmp TS_switching_pushready
org TS_STARTJMPVCT + (TS_SPACE4PANDP*3)	; Task4
	push Acc	
	push 0
	jmp TS_switching_pushready
org TS_STARTJMPVCT + (TS_SPACE4PANDP*4)	; Task5
	jmp TS_switching_pushready

org TS_STARTJMPVCT + (TS_SPACE4PANDP*5) ; Task1
	pop 0
	jmp TS_switching_popready
org TS_STARTJMPVCT + (TS_SPACE4PANDP*6)	; Task2
	pop Acc
	jmp TS_switching_popready
org TS_STARTJMPVCT + (TS_SPACE4PANDP*7)	; Task3
	pop Acc
	jmp TS_switching_popready
org TS_STARTJMPVCT + (TS_SPACE4PANDP*8)	; Task4
	pop 0
	pop Acc	
	jmp TS_switching_popready
org TS_STARTJMPVCT + (TS_SPACE4PANDP*9)	; Task5
	jmp TS_switching_popready


; Tabelle der Prioritten
;  die Werte stehen fr TL0,TH0 
TS_priority:
	db 223,177	; Task1  10ms
	db 223,177	; Task2  10ms
	db 239,216	; Task3   5ms
	db 223,177	; Task4  10ms
	db 103,197	; Task5 7.5ms

$INCLUDE (time.inc)

; Arbeit getan, nchster Task kann Prozessor haben
next_task:

; Interruptbehandlungsroutine Timer 0
;  Task-Umschaltung
timer0_int:
	; Speicher des letzten Tasks sichern
	push PSW	; PSW muss sowieso gesichert werden

	; die von der Taskumschaltung selbst genutzten Variablen sichern
	mov TS_ACCBACKUP,a
	mov TS_R0BACKUP,R0
	mov TS_DPLBACKUP,DPL
	mov TS_DPHBACKUP,DPH
	mov TS_BBACKUP,b
	
	; weitere Push's
	mov DPTR,#TS_STARTJMPVCT
	mov b,#TS_SPACE4PANDP
	mov a,TS_CURRTASK
	mul ab
	jmp @A+DPTR
	
TS_switching_pushready:
	
	; SP sichern
	mov R0,TS_CURRTASK
	mov a,#TS_DATA
	add a,R0
	mov R0,a
	mov @R0,SP

	inc R0		; Zeiger fr nchsten Task
	inc TS_CURRTASK
	
	; ist das Ende des Task-Feldes erreicht?
	cjne R0,#TS_TASKCOUNT + TS_DATA,TS_Umschaltung_L1
	mov R0,#TS_DATA	; Zhler zurcksetzen
	mov TS_CURRTASK,#0
TS_Umschaltung_L1:

	mov SP,@R0	; SP wiederherstellen

	; die gesicherten Werte vom Stack holen
	mov DPTR,#TS_STARTJMPVCT + (TS_SPACE4PANDP*TS_TASKCOUNT)
	mov b,#TS_SPACE4PANDP
	mov a,TS_CURRTASK
	mul ab
	jmp @A+DPTR

TS_switching_popready:

	; Timer erneut laden, Werte aus der Priority-Tabelle verwenden
	mov DPTR,#TS_priority	
	mov a,TS_CURRTASK
	rl a
	mov R0,a
	inc R0
	movc a,@A+DPTR
	mov TL0,a
	mov a,R0
	movc a,@A+DPTR
	mov TH0,a

	; die von der Taskumschaltung vernderten Variablen restaurieren
	mov R0,TS_R0BACKUP
	mov a,TS_ACCBACKUP
	mov DPL,TS_DPLBACKUP
	mov DPH,TS_DPHBACKUP
	mov b,TS_BBACKUP

	pop PSW
        reti

; Task-Switching abstellen
stop_switching:
	clr TCON.4
	clr IE.1
	ret

; Task-Switching wieder anstellen
start_switching:
	setb IE.1
	setb TCON.4
	ret

; Tasks initialisieren
;  Parameter: 	R0 	- Task-Nummer
;		R1	- Startadresse des Stackpointer
;		R2	- zustzliche Bytes, die dieser Task auf dem Stack belegt
; 		DPTR	- Einsprungadresse des Tasks
inittask:

	; Fr jeden Task vorbelegen:
	;  Einsprungadresse in den Task
	;  Adresse des Stackpointers fr diesen Task im TS_DATA
	;   (Anfangsadresse SP + (Anzahl gesicherter Bytes-1) + 2 fr RET-Adresse)
	;  PSW

	push 1
	mov @R1,DPL	; Einsprungadresse
	inc R1
	mov @R1,DPH
	inc R1
	mov @R1,#0	; PSW-Vorbelegung
	inc R1
	mov @R1,#0	; Acc-Vorbelegung
	inc R1
	mov @R1,#0	; R0-Vorbelegung
	pop 1

	dec R0		; Adresse des SP in TS_DATA ablegen
	mov a,#TS_DATA
	add a,R0
	mov R0,a
	
	mov a,R2
	inc a
	inc a
	add a,R1
	mov @R0,a
	ret	


; Initialisierung
;---------------------------------------------------------------
start:

; Timer 0 aktivieren
;  Software-Kontrolle
	mov DPTR,#TS_priority	; Timer laden
	clr a
	movc a,@A+DPTR
	mov TL0,a
	mov a,#1
	movc a,@A+DPTR
	mov TH0,a

; die SFR's initialisieren
        mov SP,  #TS_SP_TASK1
        mov TMOD,#1
	mov IE,  #128

	mov P1,#00100000b
	mov P3,#0

; Hauptprogramm
;---------------------------------------------------------------
main:
	; Das Task-Switching vorbereiten

	mov TS_CURRTASK,#0	; Starten mit dem 1. Task

	; Task1
	; dieser Task luft grade und wird beim nchsten
	; Task-Wechsel sowieso gesichert. Eine Vorbelegung des
	; Feldes ist daher nicht notwendig.

	; Task2
	mov DPTR,#task2		; Einsprungadresse
	mov R0,#2		; Task-Nummer
	mov R1,#TS_SP_TASK2	; SP-Start fr den Task
	mov R2,#1		; dieser Task push't 1 Byte (s. Tabelle)
	call inittask
	
	; Task3
	mov DPTR,#task3
	mov R0,#3
	mov R1,#TS_SP_TASK3
	mov R2,#1
	call inittask

	; Task4
	mov DPTR,#task4
	mov R0,#4
	mov R1,#TS_SP_TASK4
	mov R2,#2
	call inittask

	; Task5
	mov DPTR,#task5
	mov R0,#5
	mov R1,#TS_SP_TASK5
	mov R2,#0
	call inittask

	call start_switching

	jmp task1



task1:
	; Demonstration fr einen 'freiwilligen' Taskwechsel:
	;  nachdem die Aktion fertig ausgefhrt wurde, ist der nchste
	;  Task an die Reihe, um keine Prozessorzeit zu verschwenden
	inc R0
	mov P3,R0
	call next_task	; Kontrolle an den nchsten Task abgeben

	jmp task1
	
task2:
	; Demonstration eines simplen Arbeitstasks
	cpl P1.0
	
	mov a,#100
	call F_wait_m
	jmp task2

task3:
	; noch ein Arbeitstask
	; dieser Task hat die selbe Warteschleife wie
	; Task2, aber eine geringere Prioritt
	cpl P1.2

	mov a,#100
	call F_wait_m
	jmp task3

task4:
	; Demonstration der Echtzeit-Prioritt
	; In regelmigen Zeitabstnden 10x mit 150ms Delay blinken
	mov a,#1
	call F_wait_s

	call stop_switching	; Alle anderen Tasks abstellen

	mov R0,#14		; Echtzeit-Kommandos ausfhren
task4_L1:	
	mov a,#100
	call F_wait_m
	cpl P1.7
	djnz R0,task4_L1

	call start_switching	; Taskumschaltung wieder zulassen
	
	jmp task4

task5:
	; Arbeitstask: den Status von P1.5 nach P1.1 kopieren
	mov c,P1.5
	mov P1.1,c
	call next_task	; Kontrolle an den nchsten Task abgeben

	jmp task5

END
